"""Test the preprocessing module."""

import pandas as pd
from napr.utils.random import rand_list_string

import pytest

from napr.apps.coconut.terpene.preprocessing import Preprocess


@pytest.fixture
def preprocessor(data):
    return Preprocess(data=data)


def test_split_bcutDescriptor(preprocessor):
    """Test the _split_bcutDescriptor function."""
    bcutDescriptor = preprocessor.data.pop("bcutDescriptor")
    assert not preprocessor._split_bcutDescriptor()

    preprocessor.data = pd.concat([preprocessor.data, bcutDescriptor], axis=1)
    preprocessor._split_bcutDescriptor()
    for column in ["bcutDescriptor_" + str(i) for i in range(6)]:
        assert column in preprocessor.data.columns


def test_extract_tax(preprocessor):
    """Test the _extract_tax function."""
    textTaxa = preprocessor.data.pop("textTaxa")
    assert not preprocessor._extract_tax()

    preprocessor.data = pd.concat([preprocessor.data, textTaxa], axis=1)
    preprocessor._extract_tax()
    taxes = ["plants", "marine", "bacteria", "fungi"]
    for column in ["textTaxa_" + tax for tax in taxes]:
        assert column in preprocessor.data.columns


def test_encode(preprocessor, train_test_data):
    """Test the _encode function."""
    train_data, test_data = train_test_data
    assert len(train_data) == len(train_test_data[0])
    assert len(test_data) == len(train_test_data[1])

    columns = ['directParentClassification']
    df_columns = train_data[columns]
    train_data.drop(columns, axis=1, inplace=True)
    assert not preprocessor._encode(train_data, test_data)
    train_data = pd.concat([train_data, df_columns], axis=1)

    for col in columns:
        nunique_train = train_data[col].nunique()
        nunique_test = test_data[col].nunique()
        preprocessor._encode(train_data, test_data)
        assert nunique_train == train_data[col].nunique()
        assert nunique_test >= test_data[col].nunique()



"""Sharing fixtures across multiple files."""

import numpy as np
import pandas as pd
import random

from sklearn.model_selection import train_test_split

import pytest

from napr.utils.random import rand_list_string


@pytest.fixture(scope="module")
def data():
    """The data for the tests."""
    chemicalSubClass_21_3 = [
        "Diterpenoids",
        "Cholestane steroids",
        "Terpene glycosides",
        "Triterpenoids",
        "Sesquiterpenoids",
        "Monoterpenoids",
        "Eicosanoids",
        "Pregnane steroids",
        "Terpene lactones",
        "Steroidal glycosides",
        "Phosphosphingolipids",
        "Ergostane steroids",
        None,
        "Steroid esters",
        "Fatty acids and conjugates",
        "Steroid lactones",
        "Sesterterpenoids",
        "Fatty acyl glycosides",
        "Glycerophosphoethanolamines",
        "Gorgostanes and derivatives",
        "Fatty alcohols",
        "Fatty amides",
        "Fatty acyl thioesters",
        "Tetraterpenoids",
        "Oxosteroids",
        "Monoradylglycerols",
        "Steroidal alkaloids",
        "Diradylglycerols",
        "Cucurbitacins",
        "Glycerophosphocholines",
        "Glycosphingolipids",
        "Fatty acid esters",
        "Stigmastanes and derivatives",
        "Quinone and hydroquinone lipids",
        "Glycerophosphoglycerophosphoglycerols",
        "Cycloartanols and derivatives",
        "Bile acids, alcohols and derivatives",
        "Glycerol vinyl ethers",
        "Glycerophosphoinositols",
        "Steroid acids",
        "Hydroxysteroids",
        "Fatty aldehydes",
        "Lineolic acids and derivatives",
        "Triradylcglycerols",
        "Fatty alcohol esters",
        "Glycerophosphoserines",
        "Hopanoids",
        "Glycerophosphoglycerols",
        "Androstane steroids",
        "Polyprenols",
        "Glycosylglycerols",
        "Estrane steroids",
        "Sesquaterpenoids",
        "Delta-7-steroids",
        "Glycerophosphoinositol phosphates",
        "17-furanylsteroids and derivatives",
        "Ecdysteroids",
        "Furostanes and derivatives",
        "Glycerophosphoglycerophosphates",
        "Ceramides",
        "Vitamin D and derivatives",
        "Azasteroids and derivatives",
        "Retinoids",
        "Sulfated steroids",
        "Furospirostanes and derivatives",
        "Physalins and derivatives",
        "Acyltrehaloses",
        "Polyterpenoids",
        "Delta-5-steroids",
        "Glycerophosphates",
        "Isoprenoid phosphates",
        "5,6-epoxysteroids",
        "CDP-glycerols",
        "C24-propyl sterols and derivatives",
        "Polyprenylphenols",
        "Glycerol ethers",
        "Oxasteroids and derivatives",
        "Lysobisphosphatidic acids",
        "Semilysobisphosphatidic acids",
        "Glycerophosphonoethanolamines",
        "Halogenated steroids",
        "Glycero-3-pyrophosphates",
        "Phosphonosphingolipids",
        "Delta-1,4-steroids",
        "Glycerophosphonocholines",
    ]
    size = len(chemicalSubClass_21_3)

    def _rand_bcutDescriptor():
        """Generate a list of random bcutDescriptor."""
        result = []
        for _ in range(size):
            rand_list = np.random.randint(low=0, high=9, size=6).tolist()
            result.append("[" + ",".join(map(str, rand_list)) + "]")
        return result

    def _rand_textTaxa():
        """Generate a list of random textTaxa."""
        return random.choices(
            population=["plants", "marine", "bacteria", "fungi"], k=size
        )

    data = {"chemicalSubClass": chemicalSubClass_21_3}
    for column in [
        "molecular_weight",
        "alogp",
        "npl_score",
        "lipinskiRuleOf5Failures",
        "hBondAcceptorCount",
        "hBondDonorCount",
    ]:
        data[column] = np.random.rand(size).tolist()
    data["bcutDescriptor"] = _rand_bcutDescriptor()
    data["textTaxa"] = _rand_textTaxa()
    data["directParentClassification"] = rand_list_string(len_str=5, size=size)
    return pd.DataFrame(data)


@pytest.fixture(scope="module")
def dropped_columns():
    """The columns to drop."""
    return ["molecular_weight", "bcutDescriptor", "textTaxa"]


@pytest.fixture(scope="module")
def train_test_data(data, dropped_columns):
    """The training and testing data."""
    data_clean = data.drop(dropped_columns, axis=1, inplace=False)
    train_data, test_data = train_test_split(
        data_clean, test_size=0.25, random_state=777
    )
    return train_data, test_data



"""Test the evaluation of classification methods."""

import numpy as np
import pandas as pd

from sklearn.neighbors import KNeighborsClassifier

import pytest
from contextlib import nullcontext as not_raises

from napr.evaluation.classification import _scores_row, eval_classification


@pytest.mark.parametrize(
    "estimator, time, scoring, y_test, pred, expected,exception",
    [
        ("", 1, [], None, None, None, pytest.raises(ValueError)),
        ("estimator", 0, [], None, None, None, pytest.raises(ValueError)),
        (
            "estimator",
            1,
            ["accuracy"],
            None,
            None,
            None,
            pytest.raises(ValueError),
        ),
        (
            "estimator",
            1,
            ["accuracy", "precision", "recall", "f1", "conf_mat"],
            pd.Series([1, 2, 1, 3]),
            np.array([1, 2, 2, 3]),
            {
                "accuracy": 0.75,
                "precision": 0.875,
                "recall": 0.75,
                "f1": 0.75,
                "conf_mat": np.array([[1, 1, 0], [0, 1, 0], [0, 0, 1]]),
            },
            not_raises(),
        ),
    ],
)
def test_scores_row(
    estimator, time, scoring, y_test, pred, expected, exception
):
    """Test the _scores_row function."""
    with exception:
        row = _scores_row(
            estimator=estimator,
            time=time,
            y_test=y_test,
            pred=pred,
            scoring=scoring,
        )
        for score in ["accuracy", "precision", "recall", "f1"]:
            assert row[score] == expected[score]
        np.testing.assert_array_equal(row["conf_mat"], expected["conf_mat"])


@pytest.mark.parametrize(
    "estimators, X, y, test_data, scoring, exception",
    [
        ("", np.array(1), np.array(1), None, None, pytest.raises(ValueError)),
        (None, np.array(1), np.array(1), None, None, pytest.raises(ValueError)),
        (
            KNeighborsClassifier(),
            None,
            np.array(1),
            None,
            None,
            pytest.raises(ValueError),
        ),
        (
            KNeighborsClassifier(),
            np.array(1),
            None,
            None,
            None,
            pytest.raises(ValueError),
        ),
        (
            KNeighborsClassifier(),
            np.random.rand(150, 2),
            np.random.randint(0, 3, 150),
            None,
            ["accuracy", "precision", "recall", "f1", "conf_mat"],
            not_raises(),
        ),
        (
            {"test_estimator": KNeighborsClassifier()},
            np.random.rand(8, 3),
            np.random.randint(0, 2, 8),
            (np.random.rand(3, 3), np.random.randint(0, 2, 3)),
            "accuracy",
            not_raises(),
        ),
    ],
)
def test_eval_classification(estimators, X, y, test_data, scoring, exception):
    """Test the eval_classification function."""
    with exception:
        scores = eval_classification(
            estimators=estimators,
            X=X,
            y=y,
            test_data=test_data,
            scoring=scoring,
            random_state=777,
            n_jobs=-1,
        )
        if isinstance(estimators, dict):
            assert "test_estimator" in scores["estimator"]

        if scoring == "accuracy":
            assert "accuracy" in scores and "precision" not in scores
        else:
            for s in scoring:
                assert s in scores



"""Statistical calculations API."""

import pandas as pd


def percent_within(
    data: list[int | float] | pd.Series,
    interval: tuple[float, float],
    inclusive: str = "both",
) -> float:
    """Returns the percentage of data within an interval.

    Args:
        data: The data.
        interval: The interval.
        inclusive: Include boundaries: "both", "neither", "left", "right".
            Defaults to "both".
    """
    data = pd.Series(data, dtype=float)

    if data.empty:
        raise ValueError("Data cannot be empty.")

    len_within = data.between(
        min(interval), max(interval), inclusive=inclusive  # type: ignore
    ).sum()
    return 100 * len_within / len(data)



"""Test the base plotting functions."""

import matplotlib.pyplot as plt
import matplotlib as mpl

from napr.plotting._base import label_subplot, reset_plt_style, set_plt_style


def test_label_subplot():
    """Test the label_subplot function."""
    fig, ax = plt.subplots()
    label = label_subplot(fig, ax, "a")

    assert label.get_text() == "a"
    assert label.get_position() == (0, 1)


def test_plt_style():
    """Test the set_plt_style and reset_plt_style functions."""
    set_plt_style()
    assert mpl.rcParams != mpl.rcParamsDefault  # type: ignore

    reset_plt_style()
    assert mpl.rcParams == mpl.rcParamsDefault  # type: ignore

    set_plt_style("ggplot_bw")
    assert mpl.rcParams["axes.grid"] is True

    set_plt_style("dark_background")
    assert mpl.rcParams["axes.facecolor"] == "black"



"""Test the random module."""

from napr.utils.random import rand_list_string

import pytest
from contextlib import nullcontext as not_raises


@pytest.mark.parametrize(
    "letters, len_str, size, exception",
    [
        (123, 1, 1, pytest.raises(TypeError)),
        ("", 2, 2, pytest.raises(ValueError)),
        ("abc", 0, 3, pytest.raises(ValueError)),
        ("abc", 2, 0, pytest.raises(ValueError)),
        ("abc", 1, 1, not_raises()),
        ("abc", 3, 2, not_raises()),
    ],
)
def test_rand_list_string(letters, len_str, size, exception):
    """Test the rand_list_string function."""
    with exception:
        rand_list = rand_list_string(letters, len_str, size)
        assert len(rand_list) == size
        for item in rand_list:
            assert len(item) == len_str



"""Test the base data related functions."""

import pytest

from napr.data._base import download


def test_download():
    """Test the download function."""
    with pytest.raises(ValueError):
        download(url="")



"""Utility functions for plotting."""

import os

import matplotlib.pyplot as plt
from matplotlib import transforms
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.text import Text


def label_subplot(
    fig: Figure,
    ax: Axes,
    label: str,
    translate: tuple[float, float] = (0, 0),
    **kwargs
) -> Text:
    """Label a subplot.

    Args:
        fig: The figure instance.
        ax: Axis of the subplot to label.
        label: Label of the subplot.
        translate: Translation of the label position, considering the figure's
            dpi. Defaults to (0, 0).
        **kwargs: Additional keyword arguments passed to Text.

    Returns:
        A label.
    """
    trans = transforms.ScaledTranslation(
        translate[0], translate[1], fig.dpi_scale_trans
    )
    text = ax.text(
        0,
        1,
        label,
        transform=ax.transAxes + trans,
        fontsize=15,
        fontweight="bold",
        **kwargs
    )
    return text


def reset_plt_style() -> None:
    """Set matplotlib plot style to default."""
    plt.style.use("default")


def set_plt_style(style: str = "ggplot_classic") -> None:
    """Set matplotlib plot style.

    Args:
        style: Style sheet. Defaults to 'ggplot_classic'.
    """
    curr_dir = os.path.dirname(__file__)
    if style == "ggplot_classic":
        stylesheet = os.path.join(curr_dir, "ggplot_classic.mplstyle")
    elif style == "ggplot_bw":
        stylesheet = os.path.join(curr_dir, "ggplot_bw.mplstyle")
    else:
        stylesheet = style

    plt.style.use(stylesheet)



"""Functions for loading data."""

import os

import pandas as pd

from . import _base


def load_terpene(
    download: bool = False, path: str = _base.CURR_DIR, version: str = "21.3"
) -> pd.DataFrame:
    """Loading the terpene dataset.

    If download=True, the data will be downloaded first.

    Args:
        download: If the data should be downloaded or loaded from local machine.
            Defaults to False.
        path: The path to save/load the data. Defaults to current directory.
        version: Version of the data. Defaults to "21.3".

    Raises:
        ValueError: if version is not in ["21.3"].
        FileNotFoundError: if download=False and the data in path is not found.

    Returns:
        The terpene data
    """
    if version not in ["21.3"]:
        raise ValueError(f"Version {version} not supported.")

    if os.path.isdir(path):
        path = os.path.join(path, f"terpene-{version}.bz2")

    if not download:
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not found.")
    else:
        if version == "21.3":
            url = "https://drive.google.com/u/0/uc?id=1HFjVme274zL1r7Cr_0q-RrMoZebbGekJ&export=download"
            _base.download(url=url, path=path)

    data = pd.read_csv(path, index_col=0, low_memory=False, compression="infer")
    return data



"""Test the helper functions."""

import numpy as np
import pandas as pd

import pytest
from contextlib import nullcontext as not_raises

from napr.utils import all_but, split_train_test, label_encode


@pytest.mark.parametrize(
    "all, but, expected",
    [
        (["a", "b", "c"], ["b", "c"], ["a"]),
        (["a", "b", "c"], ["a", "b", "c"], []),
        (["a", "b", "c"], [], ["a", "b", "c"]),
        ([], ["a", "b", "c"], []),
        ([], [], []),
        (["a", "b", "c"], ["a", "b", "c", "d"], []),
        (["a", "b", "c"], ["d", "e", "f"], ["a", "b", "c"]),
        (["a", "b", "c"], None, ["a", "b", "c"]),
    ],
)
def test_all_but(all, but, expected):
    """Test the all_but function."""
    assert all_but(all=all, but=but) == expected


def test_split_train_test():
    """Test the split_train_test function."""
    # Without target and without selected_classes
    X_train, X_test = split_train_test(  # type: ignore
        data=pd.DataFrame(np.random.rand(10, 3), columns=["a", "b", "c"]),
        test_size=0.4,
    )
    assert X_train.shape == (6, 3)
    assert X_test.shape == (4, 3)

    # With target and without selected_classes
    X_train, X_test, y_train, y_test = split_train_test(  # type: ignore
        data=pd.DataFrame(np.random.rand(10, 3), columns=["a", "b", "c"]),
        target="c",
        test_size=0.4,
    )
    assert X_train.shape == (6, 2)
    assert y_train.shape == (6,)
    assert X_test.shape == (4, 2)
    assert y_test.shape == (4,)

    # With target and with selected_classes
    X_train, X_test, y_train, y_test = split_train_test(  # type: ignore
        data=pd.DataFrame(
            {
                "a": np.random.rand(5),
                "b": [2, 0, 1, 1, 2],
                "c": np.random.rand(5),
                "d": np.random.rand(5),
            }
        ),
        target="b",
        selected_classes=[2, 1],
        test_size=0.5,
    )
    assert X_train.shape == (2, 3)
    assert y_train.shape == (2,)
    assert X_test.shape == (2, 3)
    assert y_test.shape == (2,)
    assert y_train.isin([1, 2]).all()
    assert not y_test.isin([0]).any()

    # Different random states
    data = pd.DataFrame(np.random.rand(50, 2))
    test_size = 0.2
    random_state_1 = 12
    random_state_2 = 890
    X_train_1, X_test_1 = split_train_test(  # type: ignore
        data=data,
        test_size=test_size,
        random_state=random_state_1,
    )
    X_train_2, X_test_2 = split_train_test(  # type: ignore
        data=data,
        test_size=test_size,
        random_state=random_state_2,
    )
    assert not np.array_equal(X_train_1, X_train_2)
    assert not np.array_equal(X_test_1, X_test_2)


@pytest.mark.parametrize(
    "train, test, expected_train_encoded, expected_test_encoded, expected_labels, exception",
    [
        (
            pd.Series(dtype="float64"),
            pd.Series(1),
            None,
            None,
            None,
            pytest.raises(ValueError),
        ),
        (
            pd.Series(["a", "b", "c"]),
            pd.Series(["a", "b"]),
            np.array([0, 1, 2]),
            np.array([0, 1]),
            {"a": 0, "b": 1, "c": 2},
            not_raises(),
        ),
        (
            pd.Series(["a", "b", "c"]),
            pd.Series(["c", "a", "b"]),
            np.array([0, 1, 2]),
            np.array([2, 0, 1]),
            {"a": 0, "b": 1, "c": 2},
            not_raises(),
        ),
        (
            pd.Series(["b", "c", "e", "d"]),
            None,
            np.array([0, 1, 3, 2]),
            None,
            {"b": 0, "c": 1, "d": 2, "e": 3},
            not_raises(),
        ),
    ],
)
def test_label_encode(
    train,
    test,
    expected_train_encoded,
    expected_test_encoded,
    expected_labels,
    exception,
):
    """Test the label_encode function."""
    with exception:
        train_encoded, test_encoded, labels = label_encode(train, test)
        assert np.array_equal(train_encoded, expected_train_encoded)
        assert np.array_equal(test_encoded, expected_test_encoded)  # type: ignore
        assert labels == expected_labels



"""Napr package."""

from napr import apps, data, plotting, utils, evaluation, hyperopt

__version__ = "0.1.5"
__author__ = "Morteza Hosseini"
__license__ = "MIT"

__all__ = ["apps", "data", "plotting", "utils", "evaluation", "hyperopt"]



"""Exploratory data analysis of terpenes."""

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.patches import ConnectionPatch
import seaborn as sns

from dataclasses import dataclass

from napr.plotting import label_subplot

SUBCLASS_NAME = {
    "Monoterpenoids": "Monoterpenes",
    "Sesquiterpenoids": "Sesquiterpenes",
    "Diterpenoids": "Diterpenes",
    "Triterpenoids": "Triterpenes",
    "Terpene glycosides": "Glycosides",
    "Terpene lactones": "Lactones",
    "Sesterterpenoids": "Sesterterpenes",
    "Sesquaterpenoids": "Sesquaterpenes",
    "Polyterpenoids": "Polyterpenes",
    "other": "Other",
}


def _filter_subclasses(data: pd.DataFrame) -> pd.DataFrame:
    """Filter our sublasses of interest."""
    return data[data["chemicalSubClass"].isin(SUBCLASS_NAME.keys())]


@dataclass
class Plot:
    """Plotting submodule for the terpenes data."""

    data: pd.DataFrame

    def dist_subclass_mw_logp_nplscore(
        self, figsize: tuple[float, float] = (9, 8)
    ) -> tuple[Figure, list[Axes]]:
        """Plot the distribution of subclasses, molecular weight, logP, and
        NPL-score.

        Args:
            figsize: (Width, height) in inches. Defaults to (9, 8).

        Returns:
            Figure and Axes.
        """
        mosaic = """
        AAABBB
        AAABBB
        AAABBB
        CCDDEE
        CCDDEE"""
        fig, ax_dict = plt.subplot_mosaic(
            mosaic=mosaic, figsize=figsize
        )  # type: ignore
        ax = list(ax_dict.values())

        # Filter our subclasses of interest
        assert "chemicalSubClass" in self.data.columns
        data = _filter_subclasses(self.data)

        # Subplot a
        def _get_pies_data(data: pd.DataFrame) -> tuple[pd.Series, pd.Series]:
            data_pie = data["chemicalSubClass"].value_counts()
            threshold = 2000
            # Small pie
            data_smallpie = data_pie[data_pie <= threshold]
            data_smallpie.sort_values(inplace=True)
            # Large pie
            data_largepie = pd.concat(
                [data_pie, pd.Series({"other": data_smallpie.sum()})]
            )
            data_largepie.drop(list(data_smallpie.index), inplace=True)
            data_largepie.sort_values(inplace=True)
            return data_largepie, data_smallpie

        def _get_labels(data):
            return [
                f"{SUBCLASS_NAME[key]}\n({val})" for key, val in data.items()
            ]

        def _plot_connect_pies(ax_a, ax_b, theta_b=250):
            """theta_b is the angle of the bottom connecting point of pie b."""
            # Pie a
            center_a, r_a = ax_a.patches[0].center, ax_a.patches[0].r
            theta1_a, theta2_a = ax_a.patches[0].theta1, ax_a.patches[0].theta2
            # Pie b
            center_b, r_b = ax_b.patches[0].center, ax_b.patches[0].r
            theta1_b, theta2_b = theta_b, -theta_b
            # Connecting line
            for th_a, th_b in [(theta1_a, theta1_b), (theta2_a, theta2_b)]:
                x_a = center_a[0] + r_a * np.cos(np.pi / 180 * th_a)
                y_a = center_a[1] + r_a * np.sin(np.pi / 180 * th_a)
                x_b = center_b[0] + r_b * np.cos(np.pi / 180 * th_b)
                y_b = center_b[1] + r_b * np.sin(np.pi / 180 * th_b)
                con = ConnectionPatch(
                    xyA=(x_a, y_a),
                    xyB=(x_b, y_b),
                    coordsA="data",
                    coordsB="data",
                    axesA=ax_a,
                    axesB=ax_b,
                    linewidth=0.7,
                )
                ax_b.add_artist(con)

        # Data
        data_largepie, data_smallpie = _get_pies_data(data)
        # Large pie chart
        ax[0].pie(
            data_largepie,
            labels=_get_labels(data_largepie),
            startangle=-1.0 / 360 * data_largepie[0],
            autopct="%.0f%%",
            explode=[0.1] + [0.0] * (len(data_largepie) - 1),
            colors=sns.color_palette("Set2", n_colors=len(data_largepie)),
        )
        # Small pie chart
        ax[1].pie(
            data_smallpie,
            labels=_get_labels(data_smallpie),
            startangle=-25,
            autopct="%.0f%%",
            radius=0.55,
            textprops={"size": "smaller"},
            colors=sns.color_palette(
                "light:#66c2a5", n_colors=len(data_smallpie)
            ),
        )
        # Connect pie charts
        _plot_connect_pies(ax_a=ax[0], ax_b=ax[1])

        # Subplot b
        def _plot_hist(data, xlim, xlabel, ylabel, ax):
            num_bins = 30
            binwidth = (xlim[1] - xlim[0]) / num_bins if xlim else None
            sns.histplot(
                data=data,
                ax=ax,
                binwidth=binwidth,
                edgecolor="#40a0ff",
                color="#40a0ff",
                alpha=0.5,
                kde=False,
                linewidth=1,
            )
            ax.set(xlim=xlim, xlabel=xlabel, ylabel=ylabel)

        def _plot_table(data, bbox, ax):
            table = ax.table(
                cellText=[(i, f"{v:.1f}") for i, v in data.items()], bbox=bbox
            )
            # Format
            table.auto_set_font_size(False)
            table.set_fontsize(8.5)
            for (row, col), cell in table.get_celld().items():
                cell.set_edgecolor("#ffc17a")
                if row == 0 or col < 0:
                    cell.set_facecolor("#fff5e8")
                else:
                    row_colors = ["#fff5e8", "white"]
                    cell.set_facecolor(row_colors[row % len(row_colors)])
                if col == 0:
                    cell.set_text_props(ha="left")

        def _plot_hist_table(
            data, ax, xlim=None, xlabel=None, ylabel=None, bbox=None
        ):
            _plot_hist(
                data=data, ax=ax, xlim=xlim, xlabel=xlabel, ylabel=ylabel
            )
            _plot_table(data=data.describe()[1:], ax=ax, bbox=bbox)

        # Molecular weight
        assert "molecular_weight" in data.columns
        _plot_hist_table(
            data=data["molecular_weight"],
            ax=ax[2],
            xlim=(0, 2000),
            xlabel="Molecular weight",
            ylabel="Count",
            bbox=(0.47, 0.54, 0.5, 0.46),
        )
        # logP
        assert "alogp" in data.columns
        _plot_hist_table(
            data=data["alogp"],
            ax=ax[3],
            xlim=(-12, 20),
            xlabel="logP",
            ylabel="",
            bbox=(0.6, 0.54, 0.4, 0.46),
        )
        # Natural product-likeness score
        assert "npl_score" in data.columns
        _plot_hist_table(
            data=data["npl_score"],
            ax=ax[4],
            xlim=(-1, 4),
            xlabel="Natural product-likeness score",
            ylabel="",
            bbox=(0.08, 0.54, 0.4, 0.46),
        )

        # Label and adjust subplots
        label_subplot(fig=fig, ax=ax[0], label="a", translate=(-0.6, -0.3))
        label_subplot(fig=fig, ax=ax[2], label="b", translate=(-0.6, -0.1))
        fig.subplots_adjust(wspace=1, hspace=-0.05)

        return fig, ax

    def violin_mw_logp_nplscore(
        self, figsize: tuple[float, float] = (8.5, 3.5)
    ) -> tuple[Figure, list[Axes]]:
        """Plot the distribution of molecular weight, logP and NPL-score for
        each terpene subclass.

        Args:
            figsize: (Width, height) in inches. Defaults to (8.5, 3.5).

        Returns:
            Figure and Axes.
        """
        fig, ax = plt.subplots(1, 3, figsize=figsize, sharey=True)

        # Filter our subclasses of interest
        data = _filter_subclasses(self.data)

        # Data
        feat_subclass = "chemicalSubClass"
        assert feat_subclass in data.columns
        top_subclasses = data[feat_subclass].value_counts()[:6].index
        data_filtered = data[data[feat_subclass].isin(top_subclasses)]

        # Plot
        def _plot_violin(
            data,
            x,
            y,
            ax,
            xlim,
            xlabel,
            ylabel,
            order,
            yticklabels=[SUBCLASS_NAME[i] for i in top_subclasses],
        ):
            # Violin
            sns.violinplot(
                data=data,
                x=x,
                y=y,
                ax=ax,
                order=order,
                scale="width",
                inner=None,  # type: ignore
                palette=sns.color_palette("Dark2"),
                linewidth=1.2,
            )
            # Mean
            mean = data.groupby(y)[x].mean()
            sns.stripplot(
                x=mean,
                y=mean.index,
                ax=ax,
                order=order,
                jitter=False,
                color="white",
                marker="o",
                size=4,
            )
            # Axes
            ax.set(
                xlim=xlim, xlabel=xlabel, ylabel=ylabel, yticklabels=yticklabels
            )

        order = list(top_subclasses)
        # Molecular weight
        assert "molecular_weight" in data.columns
        _plot_violin(
            data=data_filtered,
            x="molecular_weight",
            y=feat_subclass,
            ax=ax[0],  # type: ignore
            xlim=(0, 2000),
            xlabel="Molecular weight",
            ylabel="Terpene subclass",
            order=order,
        )
        # logP
        assert "alogp" in data.columns
        _plot_violin(
            data=data_filtered,
            x="alogp",
            y=feat_subclass,
            ax=ax[1],  # type: ignore
            xlim=(-12, 20),
            xlabel="logP",
            ylabel="",
            order=order,
        )
        # Natural product-likeness score
        assert "npl_score" in data.columns
        _plot_violin(
            data=data_filtered,
            x="npl_score",
            y=feat_subclass,
            ax=ax[2],  # type: ignore
            xlim=(-1, 4),
            xlabel="Natural product-likeness score",
            ylabel="",
            order=order,
        )

        # Label and adjust subplots
        label_subplot(
            fig=fig, ax=ax[0], label="a", translate=(-1.05, 0)  # type: ignore
        )
        label_subplot(
            fig=fig, ax=ax[1], label="b", translate=(-0.25, 0)  # type: ignore
        )
        label_subplot(
            fig=fig, ax=ax[2], label="c", translate=(-0.25, 0)  # type: ignore
        )
        fig.tight_layout()

        return fig, ax  # type: ignore

    def lipinsky(
        self, figsize: tuple[float, float] = (6, 8)
    ) -> tuple[Figure, list[Axes]]:
        """Plot the Lipinsky's rule of five violations

        Args:
            figsize: (Width, height) in inches. Defaults to (6, 8).

        Returns:
            Figure and Axes.
        """
        fig = plt.figure(figsize=figsize)
        nrow, ncol = 3, 3

        # Filter our subclasses of interest
        data = _filter_subclasses(self.data)

        feat_subclass = "chemicalSubClass"
        assert feat_subclass in data.columns
        top_subclasses = data[feat_subclass].value_counts()[:6].index
        data_filtered = data[data[feat_subclass].isin(top_subclasses)]

        # Subplot a
        # Axes
        ax: list[Axes] = []
        ax.append(
            plt.subplot2grid(
                shape=(nrow, ncol), loc=(0, 0), colspan=3
            )  # type: ignore
        )
        # ax=[plt.subplot2grid(shape=(nrow, ncol), loc=(0, 0), colspan=3)]
        b_left, b_bottom, b_width, b_height = ax[-1].get_position().bounds
        ax[-1].set_position(
            [
                b_left + b_width / 10,
                b_bottom + b_height / 5,
                0.9 * b_width,
                b_height,
            ]
        )
        # Data
        feat_lipinsky = "lipinskiRuleOf5Failures"
        assert feat_lipinsky in data_filtered.columns
        counts_lipinsky = data_filtered[feat_lipinsky].value_counts()
        # Plot
        sns.countplot(
            data=data_filtered,
            x=feat_lipinsky,
            order=counts_lipinsky.index,
            edgecolor="black",
            palette=sns.color_palette("blend:white,darkred"),
            ax=ax[-1],
        )
        ax[-1].set(xlabel="", ylabel="Count")
        ax[-1].set_title(
            "Number of Lipinski's rule of five violations", loc="center"
        )
        ax[-1].bar_label(
            container=ax[-1].containers[0],
            labels=[
                f"{100 * val/len(data_filtered):.0f}%"
                for val in counts_lipinsky.values
            ],
        )

        # Subplot b
        def _plot_spider(data, ax, xticklabels, title, color):
            # Angles of starting points of sectors
            N = len(xticklabels)
            angles = [2 * np.pi * n / float(N) for n in range(N)]
            angles += angles[:1]
            # The first sector starts from the top (pi/2)
            ax.set_theta_offset(np.pi / 2)
            # Sectors direction is clockwise
            ax.set_theta_direction(-1)
            # xticks. Number of violations
            plt.xticks(
                ticks=angles[:-1],
                labels=xticklabels,
                size=9,
                position=(0.1, 0.1),
            )
            # yticks. Values for smaller circles
            ax.set_rlabel_position(-90)
            yticklabels = np.arange(20, 100, step=20)
            plt.yticks(
                ticks=[t - 1 for t in yticklabels],
                labels=[str(lbl) for lbl in yticklabels],
                color="#404040",
                size=7.5,
            )
            plt.ylim(0, 80)
            # Main plot
            ax.plot(angles, data, color=color, linewidth=1.5, linestyle="solid")
            ax.fill(angles, data, color=color, alpha=0.7)
            # Title
            plt.title(title, y=1.1, color="black", size=11, loc="center")

        # Data
        df = 100 * pd.crosstab(
            data_filtered[feat_subclass],
            data_filtered[feat_lipinsky],
            normalize="index",  # type: ignore
        )
        df = df.reindex(list(top_subclasses)).reset_index()  # type: ignore
        df[feat_subclass] = df[feat_subclass].map(SUBCLASS_NAME)
        # Plot
        colormap = plt.cm.get_cmap("Dark2", len(df))
        for i in range(len(df)):
            # Axes
            ax.append(fig.add_subplot(nrow, ncol, i + 4, polar=True))
            # Data
            values = df.loc[i].drop(feat_subclass).values.flatten().tolist()
            values += values[:1]
            # Plot
            _plot_spider(
                data=values,
                ax=ax[-1],
                xticklabels=list(df)[1:],
                title=df[feat_subclass][i],
                color=colormap(i),
            )

        # Label subplots
        label_subplot(fig=fig, ax=ax[0], label="a", translate=(-0.6, 0.15))
        label_subplot(fig=fig, ax=ax[1], label="b", translate=(-0.1, 0.27))

        return fig, ax

    def hbond(
        self, figsize: tuple[float, float] = (7, 10)
    ) -> tuple[Figure, list[Axes]]:
        """Plot the distribution of hydrogen bond acceptors and donors.

        Args:
            figsize: (Width, height) in inches. Defaults to (7, 10).

        Returns:
            Figure and Axes.
        """
        fig, ax = plt.subplots(
            7,
            2,
            figsize=figsize,
            gridspec_kw={
                "width_ratios": [1.5, 1],
                "height_ratios": [1.5, 1, 1, 1, 1, 1, 1],
            },
        )
        ax = ax.flatten()

        # Data
        # Filter our subclasses of interest
        data = _filter_subclasses(self.data)

        feat_subclass = "chemicalSubClass"
        assert feat_subclass in data.columns
        top_subclasses = data[feat_subclass].value_counts()[:6].index
        data_filtered = data[data[feat_subclass].isin(top_subclasses)]

        # Plot
        def _bar_label(ax, full_data):
            # Write 0% when there is no bar
            np.nan_to_num(ax.containers[0].datavalues, copy=False)
            ax.bar_label(
                container=ax.containers[0],
                fontsize=8.5,
                padding=1,
                labels=[
                    f"{100 * val/len(full_data):.0f}%"
                    for val in ax.containers[0].datavalues
                ],
            )

        def _plot(data, feature, upper, color, title, ylabel, ax):
            # Data
            data_filtered = data[data[feature] <= upper]
            counts = data_filtered[feature].value_counts(sort=False)
            for i in range(upper + 1):
                if i not in counts.index:
                    counts.loc[i] = 0
            counts = counts.sort_index()
            # Plot
            sns.countplot(
                data=data_filtered,
                x=feature,
                order=counts.index,
                ax=ax,
                color=color,
                edgecolor="black",
                linewidth=1,
            )
            ax.set(xlabel="", ylabel=ylabel)
            ax.set_title(title, loc="center", fontsize=10.5)
            _bar_label(ax=ax, full_data=data)

        feat_hbond_acceptor = "hBondAcceptorCount"
        feat_hbond_donor = "hBondDonorCount"
        assert feat_hbond_acceptor in data_filtered.columns
        assert feat_hbond_donor in data_filtered.columns
        color = sns.color_palette("Dark2", n_colors=8)
        for i, (feature, upper) in enumerate(
            zip([feat_hbond_acceptor, feat_hbond_donor], [10, 5])
        ):
            # Subplots a, b
            _plot(
                data=data_filtered,
                feature=feature,
                upper=upper,
                color=color[-1 - i],  # type: ignore
                title="Terpenes",
                ylabel="" if i else "Count",
                ax=ax[i],
            )
            # Subplots c, d
            for j, subclass in enumerate(top_subclasses):
                _plot(
                    data=data_filtered[
                        data_filtered[feat_subclass] == subclass
                    ],
                    feature=feature,
                    upper=upper,
                    color=color[j],  # type: ignore
                    title=SUBCLASS_NAME[subclass],
                    ylabel="" if i else "Count",
                    ax=ax[2 + i + 2 * j],
                )
        ax[12].set_xlabel("Hydrogen bond acceptor")
        ax[13].set_xlabel("Hydrogen bond donor")

        # Labels and layout
        label_subplot(fig=fig, ax=ax[0], label="a", translate=(-0.55, 0.15))
        label_subplot(fig=fig, ax=ax[1], label="b", translate=(-0.35, 0.15))
        label_subplot(fig=fig, ax=ax[2], label="c", translate=(-0.55, 0.15))
        label_subplot(fig=fig, ax=ax[3], label="d", translate=(-0.35, 0.15))
        fig.tight_layout()

        return fig, ax



"""Hyperparameter optimization."""

import numpy as np
import pandas as pd

from sklearn import metrics
from sklearn.model_selection import StratifiedKFold
from sklearn.utils.class_weight import compute_sample_weight

import keras_tuner as kt

from typing import TypeVar

EstimatorT = TypeVar("EstimatorT")


def find_best_models(
    X: pd.DataFrame | np.ndarray,
    y: pd.Series | np.ndarray,
    hypermodel: EstimatorT,
    max_trials: int = 10,
    score: str = "accuracy",
    cv: int = 5,
    directory: str = ".",
    project_name: str = "tuner",
    overwrite: bool = False,
    random_state: int | None = None,
    num_models: int = 1,
) -> EstimatorT | list[EstimatorT]:
    """Search the hyperparameters space and find the best model(s).

    Args:
        X: Input data.
        y: Target data (1d).
        hypermodel: Hyper model. Defaults to None.
        max_trials: Maximum number of trials. Defaults to 10.
        score: Score to use for hyperparameter optimization. Defaults to
            'accuracy'.
        cv: Number of folds for cross-validation. Defaults to 5.
        directory: Directory to save the tuner. Defaults to '.' (current
            directory).
        project_name: Project name for the tuner. Defaults to 'tuner'.
        overwrite: Whether to overwrite the tuner if it already exists. Defaults
            to False.
        random_state: The random state. Defaults to None.
        num_models: Number of best models found. Defaults to 1.

    Returns:
        The best model or the list of the best models.
    """
    if X.size == 0 or y.size == 0:
        raise ValueError("X and y must not be empty.")
    if hypermodel is None:
        raise ValueError("hypermodel must not be None.")

    match score:
        case "accuracy":
            scoring = metrics.make_scorer(metrics.accuracy_score)
        case "precision":
            scoring = metrics.make_scorer(
                metrics.precision_score, average="weighted"
            )
        case "recall":
            scoring = metrics.make_scorer(
                metrics.recall_score, average="weighted"
            )
        case "f1":
            scoring = metrics.make_scorer(metrics.f1_score, average="weighted")
        case _:
            raise ValueError(f"Unknown score: {score}")

    oracle = kt.oracles.BayesianOptimizationOracle(
        objective=kt.Objective("score", "max"),
        max_trials=max_trials,
    )
    cv = StratifiedKFold(cv, shuffle=True, random_state=random_state)

    tuner = kt.tuners.SklearnTuner(
        oracle=oracle,
        hypermodel=hypermodel,
        scoring=scoring,
        cv=cv,
        directory=directory,
        project_name=project_name,
        overwrite=overwrite,
    )
    sample_weight = compute_sample_weight("balanced", y)
    tuner.search(X=X, y=y, sample_weight=sample_weight)
    best_models = tuner.get_best_models(num_models=num_models)

    return best_models[0] if num_models == 1 else best_models



"""Test the base hyperopt related functions."""

from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd

from sklearn.neighbors import KNeighborsClassifier

import pytest
from contextlib import nullcontext as not_raises

from napr.hyperopt._base import find_best_models


@pytest.mark.parametrize(
    "X, y, hypermodel, scores, max_trials, cv, exception",
    [
        (
            pd.DataFrame(),
            np.array(3),
            None,
            ["accuracy"],
            1,
            2,
            pytest.raises(ValueError),
        ),
        (
            np.array(1),
            np.array(2),
            None,
            ["accuracy"],
            1,
            2,
            pytest.raises(ValueError),
        ),
        (
            np.random.rand(150, 2),
            np.random.randint(0, 2, 150),
            lambda hp: KNeighborsClassifier(hp.Choice("k", [3])),
            ["score"],
            1,
            3,
            pytest.raises(ValueError),
        ),
        (
            np.random.rand(150, 2),
            np.random.randint(0, 3, 150),
            lambda hp: KNeighborsClassifier(
                hp.Choice(name="n_neighbors", values=[3, 5])
            ),
            ["accuracy", "precision", "recall", "f1"],
            1,
            2,
            not_raises(),
        ),
    ],
)
def test_find_best_models(X, y, hypermodel, scores, max_trials, cv, exception):
    """Test the find_best_models function."""
    with exception, TemporaryDirectory() as tmpdir:
        for score in scores:
            best_model = find_best_models(
                X=X,
                y=y,
                hypermodel=hypermodel,
                score=score,
                max_trials=max_trials,
                cv=cv,
                project_name=tmpdir,
                overwrite=True,
            )
            assert isinstance(best_model, KNeighborsClassifier)
            assert best_model.n_neighbors in [3, 5]  # type: ignore



"""The base class for terpenes."""

import pandas as pd

from napr import plotting
from . import explore
from .preprocessing import Preprocess, DimReduce
from napr.utils.decorators import info


class Terpene:
    """The Terpene class."""

    def __init__(self, data: pd.DataFrame) -> None:
        self.data = data

        # Default plot style is 'ggplot_classic'
        plotting.set_plt_style()
        self.plot = explore.Plot(self.data)

    @info(message="Data preprocessing")
    def preprocess(self, **kwargs) -> None:
        """Preprocessing the terpene data.

        Args:
            **kwargs:
                random_state: The random state for train_test_split. Defaults to
                    777.
                unknown_value: The unknown value for encoding. Defaults to 9999.
                dropped_columns: The dropped columns.
        """
        preprocessor = Preprocess(data=self.data)
        self.data = preprocessor.preprocess(**kwargs)

    @info(message="Dimension reduction")
    def dim_reduce(self, inplace: bool = True, **kwargs) -> pd.DataFrame | None:
        """Dimension reduction of the terpene data.

        Args:
            inplace: Whether to perform the dimension reduction inplace.
            **kwargs:
                model: The model to use for dimension reduction. Defaults to PCA
                    keeping 95% of the variance.

        Returns:
            The dimension reduced data, or None if inplace is True.
        """
        data_reduced = DimReduce(data=self.data).dim_reduce(**kwargs)
        if inplace:
            self.data = data_reduced
        else:
            return data_reduced



"""Test the Statistical calculation functions."""

import pandas as pd

import pytest
from contextlib import nullcontext as not_raises

from napr.utils import percent_within


@pytest.mark.parametrize(
    "data, interval, inclusive, expected, exception",
    [
        ([], (2, 5), "both", None, pytest.raises(ValueError)),
        ([2, 3, 5, 7.0, 8], (3, 5), "both", 40, not_raises()),
        (pd.Series([2, 3, 5, 8.6]), (3.00001, 8), "left", 25.0, not_raises()),
        ([2], (-1, 1), "right", 0, not_raises()),
        (pd.Series([3.00, 4, 5, 7]), (1e0, 9.2), "neither", 100, not_raises()),
    ],
)
def test_percent_within(data, interval, inclusive, expected, exception):
    """Test the percent_within function."""
    with exception:
        assert expected == percent_within(
            data=data, interval=interval, inclusive=inclusive
        )



"""Generate random numbers and strings."""

import random
import string


def rand_list_string(
    letters: str = string.ascii_lowercase, len_str: int = 1, size: int = 1
) -> list[str]:
    """Returns a list of random strings.

    Args:
        letters: The letters to use in the random strings. Defaults to
            string.ascii_lowercase.
        len_str: The length of the strings. Defaults to 1.
        size: The size of the list. Defaults to 1.

    Raises:
        TypeError: if letters is not a string.
        ValueError: if letters is empty.
        ValueError: if len_str is less than 1.
        ValueError: if size is less than 1.
    """
    if not isinstance(letters, str):
        raise TypeError("letters must be a string.")
    if not letters:
        raise ValueError("letters must not be empty.")
    if len_str < 1:
        raise ValueError("len_str must be greater than 0.")
    if size < 1:
        raise ValueError("size must be greater than 0.")

    return [
        "".join(random.choices(population=letters, k=len_str))
        for _ in range(size)
    ]



"""Test for decorators."""

from napr.utils.decorators import info


def test_info():
    """Test the info decorator."""

    @info(message="Test message")
    def test_func():
        """Test function."""
        pass

    test_func()
    assert test_func.__name__ == "test_func"
    assert test_func.__doc__ == "Test function."
    assert test_func.__wrapped__.__name__ == "test_func"
    assert test_func.__wrapped__.__doc__ == "Test function."



"""Helper functions."""

import numpy as np
import pandas as pd

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder


def all_but(all: list[str], but: list[str] | None) -> list[str]:
    """Returns all items in 'all' but the items in 'but'.

    Args:
        all: All items.
        but: Items to be excluded.
    """
    if not but:
        return all

    return [item for item in all if item not in but]


def split_train_test(
    data: pd.DataFrame,
    target: str | None = None,
    selected_classes: list[str] | list[int] | None = None,
    test_size: float | None = None,
    random_state: int | None = None,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series] | tuple[
    pd.DataFrame, pd.DataFrame
]:
    """Split the data into train and test sets.

    It currently supports a 1d target.

    Args:
        data: The data to be split.
        target: The target column. Defaults to None.
        selected_classes: The classes of interest in the target. Defaults to
            None.
        test_size: Fraction of the test data. Defaults to None.
        random_state: Random state. Defaults to None.

    Returns:
        The train and test data.
    """
    if target and selected_classes:
        X_train = data[data[target].isin(selected_classes)]
    else:
        X_train = data.copy()

    if not target:
        X_train, X_test = train_test_split(
            X_train, test_size=test_size, random_state=random_state
        )
        return X_train, X_test

    y_train = X_train.pop(target)
    X_train, X_test, y_train, y_test = train_test_split(
        X_train, y_train, test_size=test_size, random_state=random_state
    )
    return X_train, X_test, y_train, y_test


def label_encode(
    train: pd.Series, test: pd.Series | None
) -> tuple[np.ndarray, np.ndarray | None, dict[str, int]]:
    """Encode 1d train and test data with labal encoder.

    Args:
        train: The train data.
        test: The test data. If None, the train data will be encoded.

    Returns:
        The encoded train (and test) data along with the labels, which shows the
            mapping of the original labels to the encoded labels.
    """
    if train.size == 0:
        raise ValueError("No train data provided.")

    # Train data
    label_encoder = LabelEncoder()
    train_encoded = label_encoder.fit_transform(train.values.ravel())

    # Test data
    if test is None or test.size == 0:
        test_encoded = None
    else:
        test_encoded = label_encoder.transform(test.values.ravel())

    # Labels
    labels = dict(
        zip(
            label_encoder.classes_,
            label_encoder.transform(label_encoder.classes_),
        )
    )
    return train_encoded, test_encoded, labels



"""Test the exploratory data analysis-related functions for terpenes."""

import random
import pandas as pd

from matplotlib import figure, axes

import pytest

from napr.apps.coconut.terpene.explore import (
    _filter_subclasses,
    SUBCLASS_NAME,
    Plot,
)


def test_filter_subclasses(data):
    """Test the filter_subclasses function."""
    filtered = _filter_subclasses(data)
    unique_subclasses = filtered["chemicalSubClass"].unique()
    subclasses = [key for key in SUBCLASS_NAME.keys() if key != "other"]
    assert set(unique_subclasses) == set(subclasses)


def test_plot_dist_subclass_mw_logp_nplscore(data):
    """Test the plot_dist_subclass_mw_logp_nplscore function."""
    plot = Plot(data)
    fig, ax = plot.dist_subclass_mw_logp_nplscore()

    assert isinstance(fig, figure.Figure)
    for axis in ax:
        assert isinstance(axis, axes.Axes)

    assert len(ax) == 5


def test_plot_violin_mw_logp_nplscore(data):
    """Test the plot_violin_mw_logp_nplscore function."""
    plot = Plot(data)
    fig, ax = plot.violin_mw_logp_nplscore()

    assert isinstance(fig, figure.Figure)
    for axis in ax:
        assert isinstance(axis, axes.Axes)

    assert len(ax) == 3


def test_plot_lipinsky(data):
    """Test the plot_lipinsky function."""
    plot = Plot(data)
    fig, ax = plot.lipinsky()

    assert isinstance(fig, figure.Figure)
    for axis in ax:
        assert isinstance(axis, axes.Axes)

    assert len(ax) == 7


def test_plot_hbond(data):
    """Test the plot_hbond function."""
    plot = Plot(data)
    fig, ax = plot.hbond()

    assert isinstance(fig, figure.Figure)
    for axis in ax:
        assert isinstance(axis, axes.Axes)

    assert len(ax) == 14



"""Decorators."""

import functools
import time
from datetime import timedelta


def info(_func=None, *, message: str):
    """Decorator to print a message before and after the function is called, and
    also the run time.

    Args:
        _func: Called function. Defaults to None.
        message: The message to be printed. Defaults to "".
    """

    def decor(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            print(message + "...", end="\r")
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            end_time = time.perf_counter()
            delta = timedelta(seconds=end_time - start_time)
            hours, minutes, seconds = str(delta).split(":")
            seconds = str(round(float(seconds)))
            print(f"{message} finished in {hours}h:{minutes}m:{seconds}s.")
            return result

        return wrapper

    return decor if _func is None else decor(_func)



"""Evaluation of the classification methods."""

import time
from tqdm import tqdm

import numpy as np
import pandas as pd

from sklearn.model_selection import train_test_split
from sklearn import metrics
from sklearn.utils.class_weight import compute_sample_weight

from typing import Any, TypeVar

EstimatorT = TypeVar("EstimatorT")


def _scores_row(
    estimator: str,
    time: float,
    scoring: list[str] = [],
    y_test: pd.Series | None = None,
    pred: np.ndarray | None = None,
    sample_weight: np.ndarray | None = None,
    average: str = "weighted",
    labels: list[str] | None = None,
) -> dict[str, Any]:
    """Reutrns a row of the scores."""
    if not estimator:
        raise ValueError("estimator is required.")
    if not time:
        raise ValueError(f"no fitting time is associated with {estimator}.")

    row = {}
    row["estimator"] = estimator
    row["time"] = time
    if "accuracy" in scoring:
        accuracy = metrics.accuracy_score(
            y_test, pred, sample_weight=sample_weight
        )
        row["accuracy"] = accuracy
    if "precision" in scoring:
        precision = metrics.precision_score(
            y_test,
            pred,
            sample_weight=sample_weight,
            average=average,
        )
        row["precision"] = precision
    if "recall" in scoring:
        recall = metrics.recall_score(
            y_test,
            pred,
            sample_weight=sample_weight,
            average=average,
        )
        row["recall"] = recall
    if "f1" in scoring:
        f1 = metrics.f1_score(
            y_test,
            pred,
            sample_weight=sample_weight,
            average=average,
        )
        row["f1"] = f1
    if "conf_mat" in scoring:
        conf_mat = metrics.confusion_matrix(
            y_test,
            pred,
            sample_weight=sample_weight,
            labels=labels,
            normalize=None,
        )
        row["conf_mat"] = conf_mat
    return row


def eval_classification(
    estimators: dict[str, EstimatorT] | list[EstimatorT] | EstimatorT,
    X: pd.DataFrame | np.ndarray,
    y: pd.Series | np.ndarray,
    test_data: tuple[pd.DataFrame, pd.Series] | None = None,
    test_size: float | None = None,
    scoring: list[str] | str | None = None,
    random_state: int | None = None,
    n_jobs: int | None = None,
) -> dict[str, Any]:
    """Evaluate classifiers.

    Args:
        estimators: The classifiers. It can be either of a single scikit-learn
            classification model, a list of models or a dictionary of
            classifiers with their names (string) as keys.
        X: Input data.
        y: The target data (1d).
        test_data: The test/validation data, in the form of (X_test, y_test). If
            not provided, the "test_size" will be considered to split the
            train-test. Defaults to None.
        test_size: Fraction of the test/evaluation data. It defaults to 0.2 in
            case neither of "test_data" nor "test_size" is provided. Defaults to
            None.
        scoring: A score or a list of scores. Defaults to None.
        random_state: Random state. Defaults to None.
        n_jobs: The number of jobs to run in parallel. Defaults to None.

    Returns:
        A dictionary mapping the score names to the evaluation results.
    """
    if isinstance(estimators, str):
        raise ValueError("estimator must be an sklearn model.")
    if not estimators:
        raise ValueError("estimators is required.")
    if X is None or X.size == 0:
        raise ValueError("X is required.")
    if y is None or y.size == 0:
        raise ValueError("y is required.")

    # Default metrics. If scoring is not specified, the default is used.
    METRICS = ["accuracy", "precision", "recall", "f1", "conf_mat"]

    if not scoring:
        scoring = METRICS
    elif isinstance(scoring, str):
        scoring = [scoring]

    # Either test_data or test_size will be used.
    if test_data:
        test_size = None
    elif not test_data and not test_size:
        test_size = 0.2

    # Names and model objects of the estimators
    if not isinstance(estimators, list) and not isinstance(estimators, dict):
        estimators = [estimators]

    if isinstance(estimators, list):
        estimator_names = [est.__class__.__name__ for est in estimators]
        estimators = {
            name: est for name, est in zip(estimator_names, estimators)
        }

    # Set common attributes for all estimators
    for estimator in estimators.values():
        setattr(estimator, "random_state", random_state)
        setattr(estimator, "n_jobs", n_jobs)

    # Data
    if test_data:
        X_train = X.copy()
        X_test = test_data[0].copy()
        y_train = y.copy()
        y_test = test_data[1].copy()
    else:
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=random_state
        )

    # Sample weights
    sample_weight_train = compute_sample_weight("balanced", y_train)
    sample_weight_test = compute_sample_weight("balanced", y_test)

    # Results
    scores = {key: [] for key in ["estimator", "time"] + scoring}

    progress_bar = tqdm(estimators.items())
    for name, estimator in progress_bar:
        progress_bar.set_postfix({"method": name})

        # Training
        start_train = time.perf_counter()
        if estimator.__class__.__name__ == "KNeighborsClassifier":
            estimator.fit(X_train, y_train)
        else:
            estimator.fit(X_train, y_train, sample_weight=sample_weight_train)
        end_train = time.perf_counter()

        # Prediction
        pred = estimator.predict(X_test)

        # Scores
        row = _scores_row(
            estimator=name,
            time=end_train - start_train,
            scoring=scoring,
            y_test=y_test,
            pred=pred,
            sample_weight=sample_weight_test,
            labels=estimator.classes_,
        )
        for key, val in row.items():
            scores[key].append(val)
    progress_bar.close()

    return scores



"""Test the base class of the terpene app."""

import pandas as pd

from sklearn.decomposition import PCA

import pytest

from napr.apps.coconut.terpene._base import Terpene
from napr.apps.coconut.terpene import explore


def test_base_class(data):
    """Test the Terpene class."""
    terpene = Terpene(data=data)

    assert hasattr(terpene, "data")
    assert isinstance(terpene.data, pd.DataFrame)
    assert terpene.data.shape == data.shape

    assert hasattr(terpene, "plot")
    assert isinstance(terpene.plot, explore.Plot)


def test_preprocess(data, dropped_columns):
    """Test the preprocessing method."""
    len_expanded_bcutDescriptor = 6  # bcutDescriptor is a string of 6 numbers
    len_expanded_textTaxa = 4  # textTaxa: plants, marine, bacteria, fungi

    terpene = Terpene(data=data)
    terpene.preprocess(
        random_state=777,
        unknown_value=9999,
        dropped_columns=dropped_columns,
    )

    assert terpene.data.shape == (
        data.shape[0],
        data.shape[1]
        + len_expanded_bcutDescriptor
        + len_expanded_textTaxa
        - len(dropped_columns),
    )
    assert "chemicalClass" not in terpene.data.columns


def test_dim_reduce(data, dropped_columns):
    """Test the dimension reduction method."""
    terpene = Terpene(data=data)
    terpene.preprocess(dropped_columns=dropped_columns)

    with pytest.raises(ValueError):
        terpene.dim_reduce(model="PCA")

    len_target_col = 1
    n_components = 2
    data_reduced = terpene.dim_reduce(
        model=PCA(n_components=n_components), inplace=False
    )
    assert data_reduced.shape == (data.shape[0], n_components + len_target_col)

    n_components = 3
    terpene.dim_reduce(model=PCA(n_components=n_components), inplace=True)
    assert terpene.data.shape == (data.shape[0], n_components + len_target_col)



"""Utility functions for working with data."""

import os
import re
from tqdm import tqdm
import requests

CURR_DIR = os.getcwd()


def download(url: str, path: str = CURR_DIR, chunk_size: int = 1024) -> None:
    """Download a file from a URL."""
    if not url:
        raise ValueError("url must not be empty.")

    try:
        responce = requests.get(url, stream=True)
        responce.raise_for_status()
    except requests.exceptions.RequestException as exc:
        raise SystemExit(exc) from exc

    # fmt: off
    if 'content-disposition' in responce.headers:
        dispos = responce.headers['content-disposition']
        file_name = re.findall("filename=\"(.+)\"", dispos)[0]

    # fmt: on
    if os.path.isdir(path):
        file_name = "tmp"
        file_path = os.path.join(path, file_name)
    else:
        file_name = os.path.basename(path)
        file_path = path

    file_size = int(responce.headers.get("content-length", 0))
    progress_bar = tqdm(
        desc=f"Downloading {file_name}",
        total=file_size,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    )
    with progress_bar, open(file_path, "wb") as file:
        for data in responce.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            progress_bar.update(size)



"""Test the data loading functions."""

import pytest

from napr.data._load import load_terpene


@pytest.mark.parametrize(
    "download, path, version, exception",
    [
        (False, ".", "20", pytest.raises(ValueError)),
        (False, "not_a_path", "21.3", pytest.raises(FileNotFoundError)),
    ],
)
def test_load_terpene(download, path, version, exception):
    """Test loading the terpene dataset."""
    with exception:
        load_terpene(download=download, path=path, version=version)



"""Preprocessing the terpenes data."""

import numpy as np
import pandas as pd

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OrdinalEncoder, StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.decomposition import PCA

from napr.utils import all_but

RANDOM_STATE = 777
TARGET = ["chemicalSubClass"]


class Preprocess:
    def __init__(self, data: pd.DataFrame) -> None:
        self.data = data.copy()

        # Defaults
        self.random_state = RANDOM_STATE
        self.unknown_value = 9999
        self.dropped_columns = [
            "_id",
            "coconut_id",
            "name",
            "iupac_name",
            "molecular_formula",
            "textTaxa",  # By _extract_tax()
            "bcutDescriptor",  # By _split_bcutDescriptor()
            "chemicalClass",  # Target is "chemicalSubClass"
            "chemicalSuperClass",  # Target is "chemicalSubClass"
        ]
        self.target_columns = TARGET

    def preprocess(self, **kwargs) -> pd.DataFrame:
        """Preprocessing of the terpene data.

        Args:
            **kwargs: Keyword arguments passed to preprocessing functions.

        Returns:
            The preprocessed data.
        """
        if "random_state" in kwargs:
            self.random_state = kwargs["random_state"]
        if "unknown_value" in kwargs:
            self.unknown_value = kwargs["unknown_value"]
        if "dropped_columns" in kwargs:
            self.dropped_columns = kwargs["dropped_columns"]

        self._split_bcutDescriptor()
        self._extract_tax()
        train, test = train_test_split(
            self.data,
            train_size=0.75,
            random_state=self.random_state,
        )
        self._encode(train, test)
        self._impute(train, test)
        self._feature_scale(train, test)

        self.data = pd.concat([train, test], axis=0)
        # self.data.sort_index(inplace=True)
        target_data = self.data[self.target_columns]
        self.data.drop(
            self.dropped_columns + self.target_columns, axis=1, inplace=True
        )
        self.data = pd.concat([self.data, target_data], axis=1)
        return self.data

    def _split_bcutDescriptor(self) -> None:
        """Split the bcutDescriptor column into multiple columns and concatenate
        to the data."""
        if "bcutDescriptor" not in self.data.columns:
            return None

        splitted = (
            self.data["bcutDescriptor"]
            .apply(lambda x: x[1:-1])
            .str.split(",", expand=True)
        )
        splitted.columns = ["bcutDescriptor_" + str(i) for i in range(6)]
        splitted.replace(["null", ""], np.nan, inplace=True)
        splitted = splitted.astype("float64")
        self.data = pd.concat([self.data, splitted], axis=1)

    def _extract_tax(self) -> None:
        """Extract the taxonomy from the textTaxa column and concatenate to the
        data."""
        if "textTaxa" not in self.data.columns:
            return None

        for tax in ["plants", "marine", "bacteria", "fungi"]:
            self.data.loc[:, "textTaxa_" + tax] = (
                self.data["textTaxa"].str.contains(tax).astype("int64")
            )

    def _encode(self, train: pd.DataFrame, test: pd.DataFrame) -> None:
        """Encode the data.

        Args:
            train: The training data.
            test: The testing data.
        """
        columns = ["directParentClassification"]

        for col in columns:
            if col not in train.columns:
                return None

        encoder = OrdinalEncoder(
            handle_unknown="use_encoded_value", unknown_value=self.unknown_value
        )
        train[columns] = encoder.fit_transform(train[columns])
        test[columns] = encoder.transform(test[columns])

    def _impute(self, train: pd.DataFrame, test: pd.DataFrame) -> None:
        """Impute the data.

        Args:
            train: The training data.
            test: The testing data.
        """
        ignored_columns = self.dropped_columns + self.target_columns
        columns = train.columns[~train.columns.isin(ignored_columns)]

        imputer = SimpleImputer(strategy="median")
        train[columns] = imputer.fit_transform(train[columns])
        test[columns] = imputer.transform(test[columns])

    def _feature_scale(self, train: pd.DataFrame, test: pd.DataFrame) -> None:
        """Feature scale the data.

        Args:
            train: The training data.
            test: The testing data.
        """
        ignored_columns = (
            self.dropped_columns
            + self.target_columns
            + list(train.columns[train.columns.str.contains("textTaxa")])
            + ["contains_sugar"]
        )
        columns = train.columns[~train.columns.isin(ignored_columns)]

        scaler = StandardScaler()
        train[columns] = scaler.fit_transform(train[columns])
        test[columns] = scaler.transform(test[columns])


class DimReduce:
    def __init__(self, data: pd.DataFrame) -> None:
        self.data = data.copy()

        # Defaults
        self.target_columns = TARGET
        self.model = PCA(n_components=0.95, random_state=RANDOM_STATE)

    def dim_reduce(self, **kwargs) -> pd.DataFrame:
        """Dimension reducttion of the terpene data.

        Args:
            **kwargs: Keyword arguments passed to dimeension reduction
                functions.

        Returns:
            The dimension reduced data.
        """
        if "model" in kwargs:
            if isinstance(kwargs["model"], str):
                raise ValueError(
                    "model object must have implemented fit_transform method."
                )
            self.model = kwargs["model"]

        columns = all_but(self.data.columns, self.target_columns)
        data_reduced = self.model.fit_transform(self.data[columns])
        df_reduced = pd.DataFrame(
            data=data_reduced,
            index=self.data.index,
            columns=["d" + str(i) for i in range(data_reduced.shape[1])],
        )
        self.data.drop(columns, axis=1, inplace=True)
        self.data = pd.concat([df_reduced, self.data], axis=1)
        return self.data



"""The utilities API."""

from ._stat import percent_within
from .helpers import all_but, split_train_test, label_encode
from . import decorators
from . import random

__all__ = [
    "percent_within",
    "all_but",
    "decorators",
    "split_train_test",
    "label_encode",
    "random",
]
